#include <stdio.h>
#include "huffman.h"

char code[MAX_SYMBOLS][MAX_LEN];
tree_node* root = NULL;
tree_node* queue_head = NULL;

/*function to crate new node*/
tree_node* allocate_new_node(int symbol, float freq)
{
  tree_node* p = (tree_node *) malloc(sizeof(tree_node));

  if (p != NULL)
  {
    p->left = p->right = p->parent = p->next = NULL;
    p->symbol = symbol;
    p->freq = freq;
    p->isleaf = 1;
  }

  return p;
}

/*
  insert node to priority queue
*/
void insert_into_priority_queue(tree_node* p)
{
  if (p->freq > 0) {
    tree_node* curr = NULL;
    tree_node* prev = NULL;
    printf("inserting:%c,%f\n", p->symbol, p->freq);

    if (queue_head == NULL) {
      queue_head = p;
      return;
    }

    curr = queue_head;
    /*find the position to insert*/
    while (curr != NULL && (curr->freq < p->freq)) {
      prev = curr;
      curr = curr->next;
    }

    if (curr == queue_head) { /*inset before curr*/
      p->next = curr;
      queue_head = p;
    } else { /*insert between prev and next*/
      prev->next = p;
      p->next = curr;
    }
  }
}

/*
    pop_priority_queue remove the first element of the queue
*/
tree_node* pop_priority_queue()
{
  tree_node* p = NULL;

  p = queue_head;
  if (p != NULL) {
    queue_head = queue_head->next;
    printf("popped:%c,%f\n", p->symbol, p->freq);
  }

  return p;
}

/*
    generates the string codes given the tree
*/
void generate_code(tree_node* root, int depth)
{
	unsigned char symbol;
	int len; /*length of code*/
	if (root->isleaf) {
		symbol = root->symbol;
		len = depth;
		code[symbol][len] = 0;
    tree_node * current = root;
    int i;
    for (i = len - 1; i >= 0; i--) {
      code[symbol][i] = (current == current->parent->left) ? '0' : '1';
      current = current->parent;
    }

		printf("built code:%c,%s\n", symbol, code[symbol]);
	} else {
		generate_code(root->left, depth + 1);
		generate_code(root->right, depth + 1);
	}
}

/*
    output code file
*/
void dump_code(FILE* fp)
{
	int i;
	for(i = 0; i < MAX_SYMBOLS; i++) {
		if (code[i][0]) /*non empty*/
			fprintf(fp,"%c %s\n", i, code[i]);
	}
}

/*
   compress the input file
*/
void encode(char* str, FILE* fout)
{
  while (*str) {
		fprintf(fout, "%s", code[(unsigned char) *str]);
		str++;
	}
}
/*
    free the tree
*/
void free_tree(tree_node* root)
{
  if (root == NULL) {
    return;
  }
  free_tree(root->left);
  free_tree(root->right);
  free(root);
}

void initialize_priority_queue(int characters, char initial_char, float frequencies[])
{
  int i;
  for(i = 0; i < characters; i++) {
    insert_into_priority_queue(allocate_new_node(initial_char + i, frequencies[i]));
  }
}

void build_tree(int characters)
{
  tree_node* p = NULL;
  tree_node* lc, *rc;

  /*build tree*/
  int i;
  for(i = 0; i < characters - 1; i++) {
    lc = pop_priority_queue();
    rc = pop_priority_queue();
    p = allocate_new_node(0, lc->freq + rc->freq);
    lc->parent = rc->parent = p;
    p->right = rc; p->left = lc;
    p->isleaf = 0;
    /*add the new node to the queue*/
    insert_into_priority_queue(p);
  }
}

/*calculate the frequency of characters from input file*/
void get_frequency_from_file(char *file_name)
{
  FILE* input_file = fopen(file_name, READ_MODE);
  char c;
  unsigned long long total = 0;

  while ((c = getc(input_file)) != EOF) {
    frequencies[(unsigned char) c]++;
    total++;
  }

  int i;
  for (i = 0; i < MAX_SYMBOLS; i++) {
    if (frequencies[i] > 0) {
      frequencies[i] /= total;
    }
  }

  fclose(input_file);
}

/* get the number of the elements*/
int count_elements(float frequencies[])
{
  int counter = 0;
  int i;
  for (i = 0; i < MAX_SYMBOLS; i++) {
    if (frequencies[i] > 0) {
      counter++;
    }
  }

  return counter;
}
/*encode the file*/
void encode_file(char *file_name)
{
  FILE* input_file = fopen(file_name, READ_MODE);
  FILE* encoded_file = fopen(OUT_FILE, WRITE_MODE);
  char c;
  while ((c = getc(input_file)) != EOF) {
    char string[] = {(unsigned char) c, '\0'};

    encode(string, encoded_file);
  }
  fprintf(encoded_file, "\n");

  fclose(input_file);
  fclose(encoded_file);
}
/*decode file*/
void decode(FILE* input_file, FILE* output_file)
{
	char c;
	tree_node* curr = root;
	while((c = getc(input_file)) != EOF)
	{
    if (curr->isleaf) {
      fprintf(output_file, "%c", curr->symbol);
      curr = root;
    }

    curr = (c == RIGHT) ? curr->right : curr->left;
	}
}
