#ifndef HUFFMAN_H
#define HUFFMAN_H

#include <stdio.h>

#define MAX_SYMBOLS 255
#define MAX_LEN     255
#define CODE_FILE "code.txt"
#define OUT_FILE "encoded.txt"

#define READ_MODE "r"
#define WRITE_MODE "w"
#define APPEND_MODE "a"
#define INPUT_FILE "book.txt"
#define RIGHT '1'
#define ENCODED "encoded.txt"
#define DECODED "decoded.txt"

typedef struct tree_node
{
    struct  tree_node* left;
    struct  tree_node* right;
    struct  tree_node* parent;
    struct  tree_node* next;
    float   freq;
    int     isleaf;
    char    symbol;
} tree_node;

extern char code[MAX_SYMBOLS][MAX_LEN];
extern tree_node* root; /*tree of symbols*/
extern tree_node* queue_head; /*list of current symbols*/

tree_node* allocate_new_node(int symbol, float freq);

void insert_into_priority_queue(tree_node* p);

tree_node* pop_priority_queue();

void generate_code(tree_node* root,int depth);

void dump_code(FILE* fp);

void encode(char* str, FILE* fout);

void free_tree(tree_node* root);

void initialize_priority_queue(int characters, char initial_char, float frequencies[]);

void build_tree(int characters);

extern float frequencies[MAX_SYMBOLS];

void get_frequency_from_file(char *file_name);

int count_elements(float frequencies[]);

void encode_file(char *file_name);

void decode(FILE* input_file, FILE* output_file);

#endif
